package sun.security.util;

import java.util.StringTokenizer;
import java.util.Locale;
import java.security.cert.CertificateParsingException;
import java.io.IOException;
import java.text.Normalizer;
import sun.security.x509.X500Name;
import javax.net.ssl.SNIHostName;
import java.util.Iterator;
import java.util.Collection;
import java.net.UnknownHostException;
import java.net.InetAddress;
import java.util.List;
import sun.net.util.IPAddressUtil;
import sun.security.ssl.Krb5Helper;
import java.security.Principal;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

public class HostnameChecker
{
    public static final byte TYPE_TLS = 1;
    private static final HostnameChecker INSTANCE_TLS;
    public static final byte TYPE_LDAP = 2;
    private static final HostnameChecker INSTANCE_LDAP;
    private static final int ALTNAME_DNS = 2;
    private static final int ALTNAME_IP = 7;
    private final byte checkType;
    
    private HostnameChecker(byte checkType) {
        this.checkType = checkType;
    }
    
    public static HostnameChecker getInstance(byte i) {
        if (i == 1) {
            return HostnameChecker.INSTANCE_TLS;
        }
        if (i == 2) {
            return HostnameChecker.INSTANCE_LDAP;
        }
        throw new IllegalArgumentException("Unknown check type: " + (int)i);
    }
    
    public void match(String s, X509Certificate x509Certificate) throws CertificateException {
        if (s == null) {
            throw new CertificateException("Hostname or IP address is undefined.");
        }
        if (isIpAddress(s)) {
            matchIP(s, x509Certificate);
        }
        else {
            this.matchDNS(s, x509Certificate);
        }
        return;
    }
    
    public static boolean match(String s, Principal principal) {
        return s.equalsIgnoreCase(getServerName(principal));
    }
    
    public static String getServerName(Principal principal) {
        return Krb5Helper.getPrincipalHostName(principal);
    }
    
    private static boolean isIpAddress(String s) {
        return IPAddressUtil.isIPv4LiteralAddress(s) || IPAddressUtil.isIPv6LiteralAddress(s);
    }
    
    private static void matchIP(String s, X509Certificate x509Certificate) throws CertificateException {
        Collection<List<?>> subjectAlternativeNames = x509Certificate.getSubjectAlternativeNames();
        if (subjectAlternativeNames == null) {
            throw new CertificateException("No subject alternative names present");
        }
        for (List list : subjectAlternativeNames) {
            if ((int)list.get(0) == 7) {
                String s2 = (String)list.get(1);
                if (s.equalsIgnoreCase(s2)) {
                    return;
                }
                try {
                    if (InetAddress.getByName(s).equals(InetAddress.getByName(s2))) {
                        return;
                    }
                    continue;
                }
                catch (UnknownHostException ex) {}
                catch (SecurityException ex2) {}
            }
            else {
                continue;
            }
        }
        throw new CertificateException("No subject alternative names matching IP address " + s + " found");
    }
    
    private void matchDNS(String s, X509Certificate x509Certificate) throws CertificateException {
        try {
            SNIHostName sniHostName = new SNIHostName(s);
        }
        catch (IllegalArgumentException cause) {
            throw new CertificateException("Illegal given domain name: " + s, cause);
        }
        Collection<List<?>> subjectAlternativeNames = x509Certificate.getSubjectAlternativeNames();
        if (subjectAlternativeNames != null) {
            boolean n = false;
            for (List list : subjectAlternativeNames) {
                if ((int)list.get(0) == 2) {
                    n = true;
                    if (this.isMatched(s, (String)list.get(1))) {
                        return;
                    }
                    continue;
                }
                else {
                    continue;
                }
            }
            if (!(n == 0)) {
                throw new CertificateException("No subject alternative DNS name matching " + s + " found.");
            }
        }
        DerValue mostSpecificAttribute = getSubjectX500Name(x509Certificate).findMostSpecificAttribute(X500Name.commonName_oid);
        if (mostSpecificAttribute != null) {
            try {
                String asString = mostSpecificAttribute.getAsString();
                if (!Normalizer.isNormalized(asString, Normalizer.Form.NFKC)) {
                    throw new CertificateException("Not a formal name " + asString);
                }
                if (this.isMatched(s, asString)) {
                    return;
                }
            }
            catch (IOException ex) {}
        }
        throw new CertificateException("No name matching " + s + " found");
    }
    
    public static X500Name getSubjectX500Name(X509Certificate x509Certificate) throws CertificateParsingException {
        try {
            Principal subjectDN = x509Certificate.getSubjectDN();
            if ((subjectDN instanceof X500Name)) {
                return (X500Name)subjectDN;
            }
            return new X500Name(x509Certificate.getSubjectX500Principal().getEncoded());
        }
        catch (IOException cause) {
            throw (CertificateParsingException)new CertificateParsingException().initCause(cause);
        }
    }
    
    private boolean isMatched(String s, String s2) {
        try {
            SNIHostName sniHostName = new SNIHostName(s2.replace('*', 'x'));
        }
        catch (IllegalArgumentException ex) {
            return false;
        }
        if (this.checkType == 1) {
            return matchAllWildcards(s, s2);
        }
        return this.checkType == 2 && matchLeftmostWildcard(s, s2);
    }
    
    private static boolean matchAllWildcards(String lowerCase, String lowerCase2) {
        lowerCase = lowerCase.toLowerCase(Locale.ENGLISH);
        lowerCase2 = lowerCase2.toLowerCase(Locale.ENGLISH);
        StringTokenizer stringTokenizer = new StringTokenizer(lowerCase, ".");
        StringTokenizer stringTokenizer2 = new StringTokenizer(lowerCase2, ".");
        if (stringTokenizer.countTokens() != stringTokenizer2.countTokens()) {
            return false;
        }
        while (stringTokenizer.hasMoreTokens()) {
            if (matchWildCards(stringTokenizer.nextToken(), stringTokenizer2.nextToken())) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    private static boolean matchLeftmostWildcard(String lowerCase, String lowerCase2) {
        lowerCase = lowerCase.toLowerCase(Locale.ENGLISH);
        lowerCase2 = lowerCase2.toLowerCase(Locale.ENGLISH);
        int n = lowerCase2.indexOf(".");
        int n2 = lowerCase.indexOf(".");
        if (n == -1) {
            n = lowerCase2.length();
        }
        if (n2 == -1) {
            n2 = lowerCase.length();
        }
        return matchWildCards(lowerCase.substring(0, n2), lowerCase2.substring(0, n)) && lowerCase2.substring(n).equals(lowerCase.substring(n2));
    }
    
    private static boolean matchWildCards(String substring, String substring2) {
        int i = substring2.indexOf("*");
        if (i == -1) {
            return substring.equals(substring2);
        }
        int n = 1;
        while (i != -1) {
            String substring3 = substring2.substring(0, i);
            substring2 = substring2.substring(i + 1);
            int index = substring.indexOf(substring3);
            if (index == -1 || (n != 0 && index != 0)) {
                return false;
            }
            n = 0;
            substring = substring.substring(index + substring3.length());
            i = substring2.indexOf("*");
        }
        return substring.endsWith(substring2);
    }
    
    static {
        INSTANCE_TLS = new HostnameChecker((byte)1);
        INSTANCE_LDAP = new HostnameChecker((byte)2);
    }
}
